/*
 * Adapted from....
 *
 * Remote debugging interface for 683xx via Background Debug Mode
 * needs a driver, which controls the BDM interface.
 * written by G.Magin
 * (C) 1995 Technische Universitaet Muenchen, Lehrstuhl fuer Prozessrechner
 * 
 * D.Jeff Dionne
 * (C) 1995, Dionne & Associates
 * 
 * V0.2 Enhancements:
 * Added command line arguments
 * Rewritten flash programming functions for more flexability
 *
 * Andrew Dennison <adennison@swin.edu.au>
 * December 1995

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 */

/*
 * NOTE:
 * This file is assumed to be runnable only on Linux/i386 because of
 * HW-restrictions: the driver is currently only available on i386-Linux 
 * So byte-swapping and alignment is handled directly, violating the 
 * GNU-coding standards. However, the "dirty spots" are restricted in this 
 * file.
 *
 * If anybody wants to port to e.g. Sparc, byte-sex has to be handled in
 * a more general way.
 */

/* 
 * the following line was added to allow bfd.h to compile on my system
 * I don't know if there is a magical compiler switch that makes this 
 * unnecessary as it wasn't in the original source. This is related to 
 * prototyping of functions.
 * AD. 26 Oct 95
 */
#ifndef PARAMS
#define PARAMS(x) x
#endif

#include <bfd.h>
#include <unistd.h>
#include <stdlib.h>
#include <varargs.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include "bdm.h"

/*
 * Global Defines
 ****************
 */

/* uncomment the following line if you want debugging messages */
#define DEBUG_MESSAGES
#define error(x) puts(x) 
#define	BDM_DEBUG_NAME	"bdm-dbg.log"
/* default delay for interface */
#define	BDM_DEFAULT_DELAY	300
#define	BDM_DODEBUG		0
//(1<<0)

/*
 * give target time to come up after reset
 * time in usec
 */
#define	TIME_TO_COME_UP	60000

/*
 * Global Types
 **************
 */
typedef int CORE_ADDR;
#ifndef BOOL
typedef enum
{
	FALSE = 0,
	TRUE
} BOOL;
#endif

/*
 * Global Variables
 ******************
 */
static long entry_pt;
static int bdm_fd;
static int bdm_delay;
char initname[255];
static char *bdm_prog_loaded;
static char hashmark;				/* flag set by "set hash" */
static int bdm_flags;
static BOOL verify = FALSE;

/*
 * FLASH Programming Variables
 *****************************
 */
u_short Flash_Width;
CORE_ADDR Flash_Base_Offset,            /* Address where FLASH starts */
	Flash_Write_Offset,             /* non-zero if R/!W on address line */
	Flash_Size,                     /* size in bytes */
	Flash_Reg1,                     /* Flash control register 1 */
	Flash_Reg2;                     /* Flash control register 2 */

/*
 * Function Prototypes
 *********************
 */
static void bdm_load(char *);
static void bdm_close();
static void bdm_fetch_register(int);
static void bdm_store_register(int);
static int bdm_get_sys_reg(int, u_int *);
static int bdm_set_sys_reg(int, u_int);
static int bdm_get_reg(int, u_int *);
static int bdm_set_reg(int, u_int);
static int bdm_do_load_binary(char *file_name);
void bdm_write_flash_block(CORE_ADDR addr,char *buf,unsigned int len);
BOOL Flash_Blank (u_long Size);
void bdm_write_flash(CORE_ADDR addr,u_int val);
BOOL bdm_erase_flash();
BOOL Flash_Blank (u_long Size);


static void bdm_check_fd()
{
	if (!bdm_fd) {
		error("\
Bdm is not yet opened. Use the 'target bdm <bdm-device>' command to open \n\
the bdm interface\n\
Use 'help bdm' for more online info");
	}
}

/* filter to pick out bdm specific error messages */
static int bdm_error(int err)
{
int ret = 0;
	switch (err) {
	case BDM_FAULT_UNKNOWN:
		fprintf(stderr, "bdm driver recognized unknown fault\n");
		break;
	case BDM_FAULT_POWER:
		fprintf(stderr, "target power failed\n");
		break;
	case BDM_FAULT_CABLE:
		fprintf(stderr, "cable disconnected\n");
		break;
	case BDM_FAULT_RESPONSE:
		fprintf(stderr, "no response from target via bdm\n");
		break;
	case BDM_FAULT_RESET:
		fprintf(stderr, "target got a reset\n");
		break;
	case BDM_FAULT_PORT:
		fprintf(stderr, "wrong bdm port\n");
		break;
	case BDM_FAULT_BERR:
		fprintf(stderr, "bus error occured on access via bdm\n");
		break;
	case BDM_FAULT_NVC:
		fprintf(stderr, 
				"bdm internal error: no valid command to bdm\n");
		break;
	default:
		ret = err;
	}
	return ret;
}

static void
bdm_setdelay(int delay)
{
int ret;
  /*
   * set delay factor for delay on bdm-interface
   */
  bdm_check_fd();
  if ((ret = ioctl(bdm_fd, BDM_SPEED, (u_long) delay))) {
  	bdm_error(ret);
  	printf("bdm_setdelay: %d\n", delay);
  }
}

/*
 * initialize bdm interface port
 */
static void
bdm_init()
{
int ret;
  bdm_check_fd();
  if ((ret = ioctl(bdm_fd, BDM_INIT, NULL))) {
  	bdm_error(ret);
  	printf("bdm_init\n");
  }
}

/*
 * get status of interface
 */
static u_short
bdm_get_status(void)
{
int ret;
  bdm_check_fd();
  ret = ioctl(bdm_fd, BDM_GET_STATUS, NULL);
  bdm_error(ret);
  return ret;
}

/*
 * restart chip: reset and go into bdm mode
 */
static void
bdm_restart_chip(void)
{
int ret;
  bdm_check_fd();
  if ((ret = ioctl(bdm_fd, BDM_RESTART_CHIP, NULL))) {
  	bdm_error(ret);
  	printf("bdm_restart_chip failed\n");
  }
  printf("bdm_restart_chip\n");
}

static void
bdm_release_chip(void)
{
int ret;

  /*
   * dont perform a bdm_check_fd() as we need to be able to quit gdb
   * if we accidentially entered bdm mode
   */
  if (bdm_fd == 0) return; 
  if ((ret = ioctl(bdm_fd, BDM_RELEASE_CHIP, NULL))) {
	fprintf(stderr, "bdm_release_chip: ");
	bdm_error(ret);
  }
}

/*
 * restart chip: reset and go into bdm mode
 */
static void
bdm_stop_chip(void)
{
int ret;

  bdm_check_fd();
  if ((ret = ioctl(bdm_fd, BDM_STOP_CHIP, NULL))) {
  	bdm_error(ret);
  }
}

static void
bdm_go(void)
{
u_short buf;
int res;
	printf("bdm_go\n");
	bdm_check_fd();
	buf = BDM_GO_CMD;
	res = write(bdm_fd, &buf, 2);
	if (res < 0) {
		bdm_error(res);
	}
}

#define	BEGIN_MACRO	1
#define	END_MACRO	0
static int bdm_do_load_macro(char *file_name, int is_begin_macro);

static void 
bdm_reset(void)
{
  bdm_restart_chip();
#if 0
  bdm_go();
  usleep(TIME_TO_COME_UP);
#endif
  bdm_do_load_macro("reset",BEGIN_MACRO);
}

/*
 * step chip: execute a single instruction
 * DJD ---Busted.  taken out as I cannot seem to fix it no matter what :-(
 */

/*
 * true if runnable
 */

#define swaps(x) \
	((u_short)((((u_short)(x) & 0x00ff) << 8) | \
		(((u_short)(x) & 0xff00) >> 8))) 

#define	swapl(x) \
	((u_int)((((u_int)(x) & 0x000000ffU) << 24) | \
		(((u_int)(x) & 0x0000ff00U) << 8) | \
		(((u_int)(x) & 0x00ff0000U) >> 8) | \
		(((u_int)(x) & 0xff000000U) >> 24)))

static u_short *
bdm_conv_short_to_buf(u_short *buf, u_short val, int endianness)
{
	if (endianness) {
		*buf++ = swaps(val);
	} else {
		*buf++ = val;
	}
	return buf;
}

/* conv long in host format (little endian) to buf (target format = big e) */
static u_short *
bdm_conv_long_to_buf(u_short *buf, u_long val, int endianness)
{
	if (endianness) {
		*buf++ = swaps(val & 0xffff);
		*buf++ = swaps(val >> 16);
	} else {
		*buf++ = val >> 16;
		*buf++ = val & 0xffff;
	}
	return buf;
}

/* buf representation to a single short */
static u_short *
bdm_conv_buf_to_short(u_short *buf, u_short *val, int endianness)
{
	if (endianness) {
		*val = *buf++;
	} else {
		*val = swaps(*buf);
		buf++;
	}
	return buf;
}

/* buf representation to a single long */
static u_short *
bdm_conv_buf_to_long(u_short *buf, u_long *val, int endianness)
{
	if (endianness) {
		*val = (buf[0] << 16)| buf[1];
	} else {
		*val = (swaps(buf[1]) << 16) | swaps(buf[0]);
	}
	buf += 2;
	return buf;
}

/*
 * put a char stream of 4 bytes into a long; 
 * endianness=0: interprete char as little end 
 * else interprete char as big endian
 */
static u_char *
bdm_conv_char_to_long(u_char *buf, u_long *val, int endianness)
{
int i;
	union {
		u_long l;
		u_char c[4];
	} u;
	if (endianness) {
		for (i=3; i >= 0; i--) u.c[i] = *buf++;
	} else {
		for (i=0; i < 4; i++) u.c[i] = *buf++;
	}
	*val = u.l;
	return buf;
}

/* put a char stream of 2 bytes into a short */
static u_char *
bdm_conv_char_to_short(u_char *buf, u_short *val, int endianness)
{
int i;
	union {
		u_short	s;
		u_char	c[2];
	} u;
	if (endianness) {
		for (i=1; i >= 0; i--) u.c[i] = *buf++;
	} else {
		for (i=0; i < 2; i++) u.c[i] = *buf++;
	}
	*val = u.s;
	return buf;
}

/* put a long to a char stream */
static u_char *
bdm_conv_long_to_char(u_char *buf, u_long val, int endianness)
{
int i;
	union {
		u_long l;
		u_char c[4];
	} u;
	u.l = val;
	if (endianness) {
		for (i=3; i >= 0; i--) *buf++ = u.c[i];
	} else {
		for (i=0; i < 4; i++) *buf++ = u.c[i];
	}
	return buf;
}

/* put a short to a char stream */
static u_char *
bdm_conv_short_to_char(u_char *buf, u_short val, int endianness)
{
int i;
	union {
		u_short	s;
		u_char	c[2];
	} u;
	u.s = val;
	if (endianness) {
		for (i=1; i >= 0; i--) *buf++ = u.c[i];
	} else {
		for (i=0; i < 2; i++) *buf++ = u.c[i];
	}
	return buf;
}

static int
bdm_get_sys_reg(int reg, u_int *ret_val)
{
u_short send;
u_short recv[2];
int res;
	send = BDM_RSREG_CMD|(reg&0xf);
  bdm_check_fd();
	res = write(bdm_fd, &send, 2);
	if (res == 2) {
		res = read(bdm_fd, recv, 4);
	} else {
		printf("bdm_get_sys_reg error on send: res %d\n", res);
		return res;
	}
	bdm_conv_buf_to_long(recv, (u_long *) ret_val, 0);
	return (res == 4) ? 0 : -res;
}

static int
bdm_set_sys_reg(int reg, u_int cont)
{
u_short send[3];
int res;

  bdm_check_fd();
	send[0] = BDM_WSREG_CMD|(reg&0xf);
	bdm_conv_long_to_buf(&send[1], cont, 0);
	if ((res = write(bdm_fd, send, 6)) != 6) {
		printf("bdm_set_sys_reg error: reg %#x cont %#x error res %#x\n", 
			reg, cont, res);
		return res;
	}
	return 0;
}

static int
bdm_get_reg(int reg, u_int *ret_val)
{
u_short send;
u_short recv[2];
int res;
  bdm_check_fd();
	send = BDM_RREG_CMD|(reg&0xf);
	res = write(bdm_fd, &send, 2);
	if (res == 2) {
		res = read(bdm_fd, recv, 4);
	} else {
		printf("bdm_get_reg: error on send res %d\n", res);
		return -res;
	}
	/* get target format, as conversion will be done in higher levels */
	bdm_conv_buf_to_long(recv, (u_long *) ret_val, 0);
	return (res==4) ? 0 : -res;
}

static int
bdm_set_reg(int reg, u_int cont)
{
u_short send[3];
int res;

  bdm_check_fd();
	send[0] = BDM_WREG_CMD|(reg&0xf);
	bdm_conv_long_to_buf(&send[1], cont, 0);
	if ((res = write(bdm_fd, send, 6)) != 6) {
		printf("bdm_set_reg error: reg %#x to cont %#x error res %#x\n", 
			reg, cont, res);
		return res;
	}
	return 0;
}

static struct _spec_regs {
	char * name;
	int	num;
} spec_regs[] = {
	{ "rpc",	BDM_REG_RPC },
	{ "pcc",	BDM_REG_PCC },
	{ "sr",	BDM_REG_SR },
	{ "usp",	BDM_REG_USP },
	{ "ssp",	BDM_REG_SSP },
	{ "sfc",	BDM_REG_SFC },
	{ "dfc",	BDM_REG_DFC },
	{ "atemp",	BDM_REG_ATEMP },
	{ "far",	BDM_REG_FAR },
	{ "vbr",	BDM_REG_VBR }
};
static int spec_regs_len = sizeof(spec_regs)/sizeof(struct _spec_regs);

static char * bdm_get_special_reg_num(char *str, int *regnum)
{
char reg_name[32];
char *s, *d;
int i;

	for (s = str, d = reg_name, i = 0; 
			(*s) && (isalpha (*s)) && i < 32;
			i++) {
		*d++ = tolower(*s++);
	}
	*d = '\0';
	for (i = 0; i < spec_regs_len; i++) {
		if (!strcmp(reg_name, spec_regs[i].name)) {
			*regnum = i;
			return s;
		}
	}
	*regnum = -1;
	return NULL;
}

static char * bdm_get_num(char *str, u_int *num)
{
char *d;
char numstr[32];
int i;

	while (isspace(*str) || (*str == '=')) str++;
	for (d = numstr, i = 0; (*str && i < 32); i++) {
		*d++ = *str++;
	}
	*d = '\0';
	if (sscanf(numstr, "%i", num) == 1) {
		return str;
	}
	*num = -1; 
	return NULL;
}

static void runerase(void)
{
int i;
	bdm_reset();
	for (i=0;i<3;i++) {
		fputs((i?"Retrying erase...":"Erasing Flash..."),stdout); fflush(stdout);
		if (bdm_erase_flash() == TRUE) {
			printf ("done.\n");
			break;
			}
		else
			printf ("Erase Failed\n");
		}
	if (i<3 && verify == TRUE) {
		printf ("Flash blank test");
		fflush(stdout);
		if (Flash_Blank (Flash_Size) == FALSE)
			puts ("Flash Contains Data");
		else
			puts ("Flash is Blank");
	}
}

static int
bdm_write_mem(CORE_ADDR adr, u_short size, u_int val)
{
u_short buf[6], *b_ptr = buf;
int buf_len;
int written;

	bdm_check_fd();
	size &= 0x00c0;
	*b_ptr++ = (BDM_WRITE_CMD | size);
	/* no need for byte swapping, as adr is already a valid long; so use 0 */
	b_ptr = bdm_conv_long_to_buf(b_ptr, adr, 0);
	switch (size) {
		case BDM_SIZE_LONG:
			b_ptr = bdm_conv_long_to_buf(b_ptr, val, 0);
			buf_len = 5;
			break;
		case BDM_SIZE_WORD:
			b_ptr = bdm_conv_short_to_buf(b_ptr, val, 0);
			buf_len = 4;
			break;
		case BDM_SIZE_BYTE:
			*b_ptr++ = (u_short)(val & 0xff);
			buf_len = 4;
			break;
		default:
			buf_len = 0;
			printf("error! bdm_write_mem: unknown size %#x\n", size);
	}
	if ((written = write(bdm_fd, buf, buf_len*2)) != buf_len*2) {
		printf("error! bdm_write_mem: write returns %d\n", written);
	}
	if ((buf_len = read(bdm_fd, buf, 2)) == 2) {
		if (*buf == 1) {
			written = -1;	/* buserr */
			printf("bdm_write_mem: buserr occured (res %#x)\n", *buf);
		}
	} else {
		printf("bdm_write_mem: verify read return val expected 2 is %d\n",
			buf_len);
	}
	return written;
}

static int
bdm_write_mem_block(CORE_ADDR in_adr, u_char *block, u_int size) {
u_short *buf, *buf_ptr;
int buf_len = 0;	/* valid len of buf to writecmd in shorts */
u_char *bl_ptr;
u_int adr;
int fills, put_size;
u_long ul; u_short us;

	bdm_check_fd();
	if (size < 4) {
		fprintf(stderr,"internal error: bdm_write_mem_block size < 4; use bdm_write_mem direct");
		return 0;
	}
	put_size = size;
	adr = (u_int) in_adr;
	bl_ptr = block;
	buf_ptr = buf = malloc((size / 4 + 5) * 6);
		/*
		 * per 4 bytes 3 shorts for fill-long-cmd
		 * + 3 short for initial write-long-cmd + addr
		 *  = 1 * 3 shorts
		 * + worst case both a write-byte and write-short on begin & end
		 *  = 4 * 3 shorts
		 */

	switch(adr & 0x3) {
		case	0:
			*buf_ptr++ = (BDM_WRITE_CMD | BDM_SIZE_LONG);
			*buf_ptr++ = (adr & 0xffff0000) >> 16;	/* h-Address */
			*buf_ptr++ = (adr & 0x0000ffff);			/* l-Address */
			bl_ptr = bdm_conv_char_to_long(bl_ptr, &ul, 1);
			buf_ptr = bdm_conv_long_to_buf(buf_ptr, ul, 0);
			adr += 4; size -= 4; buf_len = 5;
			break;
		case	1:
			*buf_ptr++ = (BDM_WRITE_CMD | BDM_SIZE_BYTE);
			*buf_ptr++ = (adr & 0xffff0000) >> 16;	/* h-Address */
			*buf_ptr++ = (adr & 0x0000ffff);			/* l-Address */
			*buf_ptr++ = *bl_ptr++;
			adr += 1; size -= 1; buf_len = 4;
			/* fall through to 'word' */
		case	2:
			*buf_ptr++ = (BDM_WRITE_CMD | BDM_SIZE_WORD);
			*buf_ptr++ = (adr & 0xffff0000) >> 16;	/* h-Address */
			*buf_ptr++ = (adr & 0x0000ffff);			/* l-Address */
			bl_ptr = bdm_conv_char_to_short(bl_ptr, &us, 1);
			buf_ptr = bdm_conv_short_to_buf(buf_ptr, us, 0);
			adr += 2; size -= 2; buf_len += 4;
			break;
		case	3:
			*buf_ptr++ = (BDM_WRITE_CMD | BDM_SIZE_BYTE);
			*buf_ptr++ = (adr & 0xffff0000) >> 16;	/* h-Address */
			*buf_ptr++ = (adr & 0x0000ffff);			/* l-Address */
			*buf_ptr++ = *bl_ptr++;
			adr += 1; size -= 1; buf_len = 4;
			break;
		default:
			/* cannot happen */
	}
	fills = size/4;
	buf_len += (fills * 3);
	while (fills--) {
		*buf_ptr++ = (BDM_FILL_CMD | BDM_SIZE_LONG);
		bl_ptr = bdm_conv_char_to_long(bl_ptr, &ul, 1);
		buf_ptr = bdm_conv_long_to_buf(buf_ptr, ul, 0);
	}

	switch(size%4) {
		case 0:
			/* nothing to be done */
			break;
		case 2:
			*buf_ptr++ = (BDM_FILL_CMD | BDM_SIZE_WORD);
			bl_ptr = bdm_conv_char_to_short(bl_ptr, &us, 1);
			buf_ptr = bdm_conv_short_to_buf(buf_ptr, us, 0);
			buf_len += 2;
			break;
		case 3:
			*buf_ptr++ = (BDM_FILL_CMD | BDM_SIZE_WORD);
			bl_ptr = bdm_conv_char_to_short(bl_ptr, &us, 1);
			buf_ptr = bdm_conv_short_to_buf(buf_ptr, us, 0);
			buf_len += 2;
			/* fall through into single byte rest */
		case 1:
			*buf_ptr++ = (BDM_FILL_CMD | BDM_SIZE_BYTE);
			*buf_ptr++ = *bl_ptr;
			buf_len += 2;
			break;
		default:
			/* cannot happen */
	}

	/* now we have a whole buf we can send in one chunk to 'write' */

	buf_len *= 2; /* now buflen in bytes */
	if ((size = write(bdm_fd, buf, buf_len)) != buf_len) {
		printf("error! bdm_write_mem_block: write %d returns %d\n",
			buf_len, size);
		put_size = 0;
	}

	/*
	 * FIXME: needs more persistence on trying to write; analyze reason
	 * for not writing, and if possible, try again with rest of the buf
	 * give up on detected bus-err
	 */

	free(buf);
	return put_size;
}

/* return format in *val is target byte ordering */
static int
bdm_read_mem(CORE_ADDR adr, u_short size, u_int *val)
{
u_short buf[6], *b_ptr = buf;
int count;
u_char *cptr;
u_short *sptr;

	size &= 0x00c0;

	bdm_check_fd();
	*b_ptr++ = (BDM_READ_CMD | size);
	/* no need for byte swapping, as adr is already a valid long */
	*b_ptr++ = (adr & 0xffff0000) >> 16;	/* h-Address */
	*b_ptr++ = (adr & 0x0000ffff);			/* l-Address */
	if ((count = write(bdm_fd, buf, 6)) != 6) {
		printf("bdm_read_mem error: write returns %d\n", count);
		return -1;
	}
	switch (size) {
		case BDM_SIZE_LONG:
			if ((count = read(bdm_fd, buf, 4)) == 4) {
				*val = (buf[0] << 16)| buf[1];
			}
			break;
		case BDM_SIZE_WORD:
			if ((count = read(bdm_fd, buf, 2)) == 2) {
				sptr = (u_short*) val;
				*sptr = (u_short) buf[0];
			}
			break;
		case BDM_SIZE_BYTE:
			if ((count = read(bdm_fd, buf, 2)) == 2) {
				count = 1;
				cptr = (u_char*)val;
				*cptr = (u_char) (buf[0] & 0xff);
			}
			break;
		default:
			printf("error! bdm_read_mem: unknown size %#x\n", size);
	}
	val = (u_int*) ((long)val & ~3);
	return count;
}

static int
bdm_read_mem_block(CORE_ADDR in_adr, u_char *bl_ptr, u_int size) {
u_short buf[8];
u_short dump_cmd;
int dumps, got_size = 0;
u_long ul; u_short us; u_char uc;

	if (size < 4) {
		fprintf(stderr,"internal error: bdm_read_mem_block size < 4; use bdm_read_mem direct");
		return 0;
	}

	switch(in_adr & 0x3) {
		case	0:
			if (bdm_read_mem(in_adr, BDM_SIZE_LONG, (u_int *) &ul) != 4) {
				return got_size;
			}
			bl_ptr = bdm_conv_long_to_char(bl_ptr, ul, 1);
			got_size += 4;
			break;
		case	1:
			if (bdm_read_mem(in_adr, BDM_SIZE_BYTE, (u_int *) &uc) != 1) {
				return got_size;
			}
			*bl_ptr++ = uc; got_size += 1; in_adr += 1;
			/* fall through to 'word' */
		case	2:
			if (bdm_read_mem(in_adr, BDM_SIZE_WORD, (u_int *) &us) != 2) {
				return got_size;
			}
			bl_ptr = bdm_conv_short_to_char(bl_ptr, us, 1);
			got_size += 2;
			break;
		case	3:
			if (bdm_read_mem(in_adr, BDM_SIZE_BYTE, (u_int *) &uc) != 1) {
				return got_size;
			}
			*bl_ptr++ = uc; got_size += 1;
			break;
		default:
			/* cannot happen */
	}


	dump_cmd = BDM_DUMP_CMD | BDM_SIZE_LONG;
	dumps = (size - got_size) / 4;
	while (dumps--) {
		write(bdm_fd, &dump_cmd, 2);		
		if (read(bdm_fd, &buf, 4) != 4) {
			fprintf(stderr,"failing read, got_size=%d\n",got_size);
			return got_size;
		}
		bdm_conv_buf_to_long(buf, &ul, 1);
		bl_ptr = bdm_conv_long_to_char(bl_ptr, ul, 1);
		got_size += 4;
	}

	switch((size - got_size) % 4) {
		case 0:
			/* nothing to be done */
			break;
		case 2:
			dump_cmd = BDM_DUMP_CMD | BDM_SIZE_WORD; 
			write(bdm_fd, &dump_cmd, 2);
			if (read(bdm_fd, &buf, 2) != 2) {
				return got_size;
			}
			bdm_conv_buf_to_short(buf, &us, 0);
			bl_ptr = bdm_conv_short_to_char(bl_ptr, us, 1);
			got_size += 2;
			break;
		case 3:
			dump_cmd = BDM_DUMP_CMD | BDM_SIZE_WORD; 
			write(bdm_fd, &dump_cmd, 2);
			if (read(bdm_fd, &buf, 2) != 2) {
				return got_size;
			}
			bdm_conv_buf_to_short(buf, &us, 0);
			bl_ptr = bdm_conv_short_to_char(bl_ptr, us, 1);
			got_size += 2;
			/* fall through into single byte rest */
		case 1:
			dump_cmd = BDM_DUMP_CMD | BDM_SIZE_BYTE;
			write(bdm_fd, &dump_cmd, 2);
			if (read(bdm_fd, &buf, 2) != 2) {
				return got_size;
			}
			*bl_ptr++ = *buf & 0xff;
			got_size += 1;
			break;
		default:
			/* cannot happen */
	}
	return got_size;

}

static char *dev_name;
/*
 * Open a connection the target via bdm
 * name is the devicename of bdm and the filename to be used
 * used for communication.
 */
static void
bdm_open()
{
	if (bdm_fd) {
		puts ("Already open");
	}
	if ((bdm_fd = open("/dev/bdm1", O_RDWR)) < 0) {
		printf("bdm_open: error on open %d\n", bdm_fd);
		bdm_fd = 0;	/* mark unused */
		exit(1);
	}

	bdm_init(0, "");
	bdm_setdelay(BDM_DEFAULT_DELAY);
	bdm_go();		
	printf("Connected to /dev/bdm1\n");
}

static void
bdm_close ()

{
	bdm_release_chip();
	close(bdm_fd);
	bdm_fd = 0;
}

/* Read the remote registers.  */
static void
bdm_fetch_registers (void)
{
int regno;

	for (regno = 0; regno <= 15; regno++)
		bdm_fetch_register(regno);
}

/*
 * Fetch register REGNO, or all registers if REGNO is -1.
 */
static void
bdm_fetch_register(int regno)
{
int val;
	if (regno < 0)
		bdm_fetch_registers();
	else {
		if (regno < 16) {
			bdm_get_reg(regno, &val);
		} else switch (regno) {
			case 16:	/* ps */
				bdm_get_sys_reg(BDM_REG_SR, &val);
				val &= 0xf;
				break;
			case 17:	/* pc */
				bdm_get_sys_reg(BDM_REG_RPC, &val);	
				break;
			default:
				fprintf (stderr,"illegal regno in bdm_fetch_register() %d", regno);
				return;
		}
		/* you would typically do something here - DJD */
	}
}

/* Store the remote registers from the contents of the block REGS.  */
static void
bdm_store_registers(void)
{
int regno;
	for (regno = 0; regno <= 15; regno++)
  		bdm_store_register(regno);
}

/* Store register REGNO, or all if REGNO == -1.  Return errno value.  */
void
bdm_store_register(int regno)
{
u_int val, temp_val;
	if (regno == -1) {
   		bdm_store_registers();
	} else {

		/*FIXME:  val = read_register (regno); */
		if (regno < 16) {
			bdm_set_reg(regno, val);
		} else switch (regno) {
			case 16:	/* ps */
				bdm_get_sys_reg(BDM_REG_SR, &temp_val);
				temp_val &= ~0xf;
				bdm_set_sys_reg(BDM_REG_SR, temp_val | val);
				break;
			case 17:	/* pc */
				bdm_set_sys_reg(BDM_REG_RPC, val);
				break;
			default:
				fprintf(stderr,"bdm_store_register: illegal regno in bdm_store_register() %d\n", regno);
		}
	}
}

/*
 * there should be something similar within gdb, e.g. for scanning .gdbinit
 * but I dont know how to use :-(
 */
#define	LINE_LEN	256
static char *get_line(FILE * f, int *line_nr)
{
static char line_buf[LINE_LEN];
	while (!feof(f)) {
		fgets(line_buf, LINE_LEN, f);
		(*line_nr)++;
		if (line_buf[0] == '#') continue;
		if (line_buf[0] == '\n') continue;
		if (line_buf[0] == '\0') continue;
		return line_buf;
	}
	return NULL;
}

#define	BEGIN_MACRO	1
#define	END_MACRO	0
static int bdm_do_load_macro(char *file_name, int is_begin_macro)
{
char 	 m_name[256];
FILE 	*m_file;
char *lptr;
char cmd;
int line_nr, ret;
CORE_ADDR	addr1, addr2;
u_int		size;
int			size_tag;
u_char *buf;

	if (bdm_do_load_binary (initname) == 0) {
		return -1;	/* init object existed, ignore macros */
	}
	strcpy(m_name, file_name);
	if ((lptr = rindex(m_name, '.')) != 0) {
		*lptr = '\0';
	}
	strcat(m_name, (is_begin_macro == BEGIN_MACRO) ? ".bdmmb" : ".bdmme");
	if ((m_file = fopen(m_name, "r")) == NULL) {
		return -1;	 /* no macro file available, quit silently */
	}
	printf("doing macro %s ",m_name);

	if (bdm_get_status() != BDM_TARGETSTOPPED) {
		printf("\twarning: chip needs stopping\n");
		bdm_stop_chip();
	}
	line_nr = 0;
	while ((lptr = get_line(m_file, &line_nr))) {
		if (hashmark) {
			putchar('.'); fflush(stdout);
		}
		sscanf(lptr, "%c %i %i %i", 
			&cmd, (unsigned int*) &addr1, (unsigned int*) &addr2, &size);
		switch (toupper(cmd)) {
			case 'W': 
				switch (size) {
					case 1: size_tag = BDM_SIZE_BYTE; size = 8; break;
					case 2: size_tag = BDM_SIZE_WORD; size = 8; break;
					case 4: size_tag = BDM_SIZE_LONG; size = 10; break;
					default:
						fprintf(stderr,"\
Error in processing macro %s line %d 'W' command:\n\t\
size must be either 1,2,4 bytes, '%d' is not allowed",
							m_name, line_nr, size);
						continue;
				}
				if ((ret = bdm_write_mem(addr1, size_tag, addr2)) != size) {
					fprintf(stderr,"\
Error in processing macro %s line %d 'W' command:\n\t\
bdm_write size mismatch: send %d returned %d",
						m_name, line_nr, size, ret);
				}
				break;
			case 'C': 
				buf = malloc(size);
				if ((ret = bdm_read_mem_block(addr1, buf, size)) != size) {
						fprintf(stderr,"\
Error in processing macro %s line %d 'C' command:\n\t\
bdm_copy size mismatch on read from %#x: wanted %d got %d",
							m_name, line_nr, addr1, size, ret);
				} else {
					if ((ret = bdm_write_mem_block(addr2, buf, size)) != size) {
						fprintf(stderr,"\
Error in processing macro %s line %d 'C' command:\n\t\
bdm_copy size mismatch on write to %#x: wanted %d got %d",
							m_name, line_nr, addr2, size, ret);
					}
				}
				free(buf);
				break;
			case 'Z': 
				buf = malloc(size);
				memset(buf, addr2, size);
				if ((ret = bdm_write_mem_block(addr1, buf, size)) != size) {
					fprintf(stderr,"\
Error in processing macro %s line %d 'Z' command:\n\t\
bdm_set size mismatch on write to %#x: wanted %d got %d",
						m_name, line_nr, addr1, size, ret);
					
				}
				free(buf);
				break;
#if DEBUGGING_MACROS
			case 'T':
				{
				int i;
				buf = malloc(size);
				for (i=0; i< size; i++) buf[i]=i;
				bdm_write_mem_block(addr1,buf,size);
				free(buf);
				}
				break;
#endif /* DEBUGGING_MACROS */
			default:	printf("\tcmd %c unknown\n", cmd);
		}
	}
	if (hashmark) {
		putchar('\n'); fflush(stdout);
	}
	return 0;
}

#define	DEFAULT_CHUNK_SIZE	512
/* 
 * load a binary file, returns error codes
 * side effect to globals: entry_pt
 */
static int
bdm_do_load_binary(char *file_name)
{
bfd *abfd;
asection *s;
char i_buffer[DEFAULT_CHUNK_SIZE]/*,o_buffer[sizeof(i_buffer)]*/;
int load_bytes;
u_int chunk_size, addr;
int ret;

	abfd = bfd_openr(file_name, 0);
	if (!abfd) {
//		fprintf(stderr,"Unable to open file %s\n", file_name);
		return -1;
	}

	if (bfd_check_format (abfd, bfd_object) == 0) {
		fprintf(stderr,"File is not an object file\n");
		return -2;
	}

	entry_pt = (u_long) bfd_get_start_address(abfd);
#ifdef DEBUG_MESSAGES
	printf("Entry point: 0x%X\n",entry_pt);
#endif
	s = abfd->sections;
	while (s != (asection *) NULL) {
#if 0
		printf("bdm_do_load_binary:\n\tsection %s index %d\n", s->name, s->index);
		printf("\tflags %#x raw_size %d cooked_size %d\n",
			s->flags, s->_raw_size, s->_cooked_size);
		printf("\tvma %#x lma %#x output_offset %#x\n",
			s->vma, s->lma, s->output_offset);
#endif
		if ((s->flags & SEC_LOAD) && (s->_raw_size >= 4)) {
#ifdef DEBUG_MESSAGES
			printf("\tloading section %s \n", s->name);	
#endif
			if (bdm_get_status() != BDM_TARGETSTOPPED) {
				fprintf(stderr, "\twarning: chip needs stopping\n");
				bdm_stop_chip();
			}
			load_bytes = 0;
			addr = s->lma;
			chunk_size = DEFAULT_CHUNK_SIZE;
			while (load_bytes + 3 < s->_raw_size) {
				if (load_bytes + chunk_size >= s->_raw_size) {
					chunk_size = (s->_raw_size - load_bytes);
				}
				if (bfd_get_section_contents(abfd, s, 
						i_buffer, load_bytes, chunk_size) == false) {
					fprintf(stderr, "read error section %s\n", 
						s->name);
					break;
				}
				if ((addr > (Flash_Base_Offset + Flash_Size)) ||
					(addr < Flash_Base_Offset))
				{
					if ((ret = bdm_write_mem_block(addr, i_buffer, chunk_size))
					    != chunk_size) {
						fprintf(stderr, "\
write error on block write, written %d acknowledged %d\n", 
						chunk_size, ret);
						break;
					}
				} else {
					bdm_write_flash_block(addr, i_buffer, chunk_size);
				}
/*
				sleep(1);
				if ((ret=bdm_read_mem_block(addr,o_buffer,chunk_size))!=chunk_size) {
					fprintf(stderr, "flash/memory read error (addr=0x%x, want=%d, got=%d) section %s\n",
						addr,chunk_size,ret,s->name);
                                        break;
					}
				else */ {
char err=0;
unsigned char val;
int i;
					for (i=0;i<chunk_size;i++)
#if 0
						if (i_buffer[i]!=o_buffer[i])
#else
						if ((ret=bdm_read_mem(addr+i,BDM_SIZE_BYTE,(int *)&val))!=1) {
							fprintf(stderr, "flash/memory read error (addr=0x%x, want=1, got=%d) section %s\n",
								addr+i,ret,s->name);
							err=1;
							}
						else if ((unsigned char)i_buffer[i]!=(unsigned char)val)
						
#endif
							{
							fprintf(stderr,"@0x%x:%02X!=f%02X, ",addr+i,
#if 0
								(unsigned char)o_buffer[i]
#else
								val
#endif
									,(unsigned char)i_buffer[i]);
							err=1;
							}
					if (err) {
						fprintf(stderr, "!\nflash/memory compare error after write, section %s\n",s->name);
//						break;
						}
					}
				load_bytes += chunk_size;
				addr += chunk_size;
				if (hashmark) {
					putchar('.'); 
					fflush(stdout);
				}
			}
		}
		s = s->next;
	}
	if (hashmark) {
		putchar('\n'); fflush(stdout);
	}
	bfd_close(abfd);
	return 0;
}


/*
 * Load a file. 
 * This is supposed to be coff here, but bfd should handle other formats...
 *
 * use a macro for target specific manipulations, eg. setting
 * chipselect, switching memory banks, etc.
 */

static void
bdm_load (char *args)
{
	int res = 0;
	u_int sfc, dfc;
	int we_have_macro_files = 0;
	char *p;

	bdm_check_fd();

	printf("bdm_load: program name %s\n", args);

	bdm_reset();
	if (!(bdm_get_status() & BDM_TARGETSTOPPED))
		bdm_stop_chip();

	bdm_set_sys_reg(BDM_REG_DFC, 5); bdm_set_sys_reg(BDM_REG_SFC, 5);
	bdm_get_sys_reg(BDM_REG_DFC, &dfc); bdm_get_sys_reg(BDM_REG_SFC, &sfc);

	/* strip off additional (unwanted) arguments after file name */

	we_have_macro_files = !bdm_do_load_macro(args, BEGIN_MACRO);

	if ((res = bdm_do_load_binary(args)) < 0) {
		fprintf(stderr,"Download failed.");
	}

	if (we_have_macro_files) bdm_do_load_macro(args, END_MACRO);
}

/*
 * Flash blank test - Added AD. 7 Nov 1995
 *****************************************
 */
BOOL Flash_Blank (u_long Size)
{
u_int Data;
u_int i = 0;
CORE_ADDR Addr = 0;

	do {
		bdm_read_mem(Flash_Base_Offset + Addr,BDM_SIZE_LONG,&Data);
		/* check if location is blank */
		if ((Data & 0xFFFFFFFF) != 0xFFFFFFFF) {
			fprintf (stderr, "Verify error: Location %8X contains %4X\n", 
				Flash_Base_Offset + Addr, Data);
			if (hashmark) putchar('\n'); 
			return FALSE;
		}
		/* put hash mark every 1K if required */
		if (hashmark && ((i+=4) > 10240)) {
			i = 0;
			putchar('.'); 
			fflush(stdout);
		}
	} while ((Addr += 4) < Size);
	if (hashmark) putchar('\n'); 
	return TRUE;
}
/*
 * Erase either 8 or 16 bit (2x8) FLASH.
 ***************************************
 */
BOOL bdm_erase_flash()
{
int Write_Error;
u_int val;
u_short Timeout = 0;

	Write_Error = bdm_write_mem(Flash_Reg1,Flash_Width,0x0AAAA);
	Write_Error |= bdm_write_mem(Flash_Reg2,Flash_Width,0x05555);
	Write_Error |= bdm_write_mem(Flash_Reg1,Flash_Width,0x08080);
	Write_Error |= bdm_write_mem(Flash_Reg1,Flash_Width,0x0AAAA);
	Write_Error |= bdm_write_mem(Flash_Reg2,Flash_Width,0x05555);
	Write_Error |= bdm_write_mem(Flash_Reg1,Flash_Width,0x01010);

/* FIXME: make bdm_write_mem return normal error codes */
	if (Write_Error > 0) {
		/* check if erased */
		do {
			bdm_read_mem(Flash_Base_Offset,BDM_SIZE_LONG,&val);
/* FIXME: Need system speed independant timeout */
			if (Timeout++ > 10000) return FALSE;
		} while ((val & 0xffffffff) != 0xffffffff);
		return TRUE;
	}
	return FALSE;
}

/*
 * Program a flash location
 **************************
 */
void bdm_write_flash(CORE_ADDR addr,u_int val)
{
/* FIXME: return an error value */
int Write_Error;
u_int rval;
u_int timeout = 0,testit;
u_int mask;            /* used for masking valid bits in data */

	mask = (Flash_Width == BDM_SIZE_BYTE) ? 0xFF : 0xFFFF;

	/* ensure location is blank */
	bdm_read_mem(addr,Flash_Width,&rval);
	if ((rval&mask) == val) return;
	if ((rval & mask) != mask) {
		fprintf (stderr, "FLASH NOT BLANK!! Address %x contains %x (val=%x)\n", addr, (u_short)rval,(u_short)val);
		exit(1);
		return;
		}
	if (val == mask) return; /* don't need to program */
	Write_Error = bdm_write_mem(Flash_Reg1,Flash_Width,0x0AAAA);
	Write_Error |= bdm_write_mem(Flash_Reg2,Flash_Width,0x05555);
	Write_Error |= bdm_write_mem(Flash_Reg1,Flash_Width,0x0A0A0);
	Write_Error |= bdm_write_mem(Flash_Write_Offset + addr,Flash_Width,val);
/* FIXME: make bdm_write_mem return normal error codes */
	if (Write_Error > 0) {
		for (testit=0;testit<10;testit++) {
		timeout=0;
		do {
			bdm_read_mem(addr,Flash_Width,&rval);
		} while (((rval & mask) != val) && timeout++ < 150);
/* FIXME: do speed independant timeout */
		if (timeout > 149) {
			fprintf (stderr, "Timeout at address %x Expected %x but got %x\n",addr,(u_short)val,(u_short)rval);
			fflush(stdout);
			bdm_init();
		}
		if (testit&&timeout) {
			fprintf (stderr, "Failed after initial at address %x Expected %x but got %x\n",addr,(u_short)val,(u_short)rval);
			}
		}
	}
	else puts ("BDM Write error");
}

/*
 * Program a block of flash memory
 *********************************
 */
void bdm_write_flash_block(CORE_ADDR addr,char *buf,unsigned int len)
{
/* FIXME: return an error value */
unsigned int p;
u_short data;

#ifdef DEBUG_MESSAGES	
//	printf ("Programming %d bytes from Address 0x%X\n", len, addr);
#endif
	if (Flash_Width == BDM_SIZE_WORD) {
		for (p = 0; p < len; p+=2) {
			data = (buf [p] << 8) | (u_char)(buf [p+1]);
			bdm_write_flash ((CORE_ADDR)(addr + p),data);
		}
	}
	else {
		for (p = 0; p < len; p++) { 
			data = buf[p];
			bdm_write_flash ((CORE_ADDR)(addr + p),data);
		}
	}
}

/*
 * Main Program
 **************
 */
main(int argc, char *argv[])
{
int done;
int i;
BOOL load = FALSE;
BOOL erase = FALSE;
int cmd;
char fname[255]="";
char buf[255];

	/* initialise global flash control variables */
	strcpy (initname, "ram_init.srec"); /* default init file for EFI332*/
	Flash_Width = BDM_SIZE_WORD;
	Flash_Base_Offset = 0;
	Flash_Write_Offset = 0; 
	Flash_Size = 0x40000;               /* default 256K for EFI332 */

	/* Extract command line information */
	for (i = 1; i < argc; i++) {
		if(*argv[i] == '-') {
			switch (*(argv[i]+1)) {
			case 'i':
				/* Init File */
				strcpy (initname, argv[++i]);
				break;
			case 'v':
				/* Verify Erase */
				verify = TRUE;
				break;
			case 'e':
				/* Erase Flash */
				erase = TRUE;
				break;
			case 'b':
				/* Flash base address */
				if (strncmp (argv[++i], "0x", 2) == 0)  /* hex input */
					sscanf (argv[i], "0x%x", &Flash_Base_Offset);
				else
					sscanf (argv[i], "%d", &Flash_Base_Offset);
				break;
			case 'o':
				/* flash write offset */
				if (strncmp (argv[++i], "0x", 2) == 0)  /* hex input */
					sscanf (argv[i], "0x%x", &Flash_Write_Offset);
				else
					sscanf (argv[i], "%d", &Flash_Write_Offset);
				break;
			case 's':
				/* flash size */
				if (strncmp (argv[++i], "0x", 2) == 0)
					sscanf (argv[i], "0x%x", &Flash_Size);
				else
					sscanf (argv[i], "%d", &Flash_Size);
				break;
			case 'f':
				/* object file to program */
				strcpy (fname, argv[++i]);
				load = TRUE;
				break;
			case 'w':
				/* flash width - 8 or 16 bits */
				if (strncmp (argv[++i], "8", 1) == 0) {
					Flash_Width = BDM_SIZE_BYTE;
				} else if (strncmp (argv[i], "16", 2) == 0) {
					Flash_Width = BDM_SIZE_WORD;
				} else {
					printf ("Invalid command \"-w %s\"", argv[i]);
					exit (-1);
				}
				break;
			case 'h':
				printf ("Usage bdm-load [OPTIONS]\n\
\t-h        - this help information!\n\
\t-i FILE   - object file to initialise processor\n\
\t-e        - erase flash\n\
\t-v        - verify erase\n\
\t-b VALUE  - flash base address, set if flash not located at 0x0000\n\
\t-o VALUE  - flash write offset if R/!W connected to address line\n\
\t-s VALUE  - flash size in bytes\n\
\t-f FILE   - object file to download\n\
\t-w [8|16] - flash width in bits\n");
				exit(0);
				break;
			default:
				printf ("Unknown command \"%s\" Try bdm-load -h for more info.\n", argv[i]);
				exit (-1);

			}
		} else {
			printf ("Wrong useage... Try bdm-load -h for more info.\n");
			exit (-1);
		}
	}
#ifdef DEBUG_MESSAGES
	printf ("Init file is: %s\n", initname);
	printf ("Flash base address is: 0x%X\n", Flash_Base_Offset);
	printf ("Flash write offset is: 0x%X\n", Flash_Write_Offset);
	printf ("Flash size is: 0x%X\n", Flash_Size);
	printf ("Object file is: %s\n", fname);
	printf ("Flash width is %d\n", (Flash_Width == BDM_SIZE_BYTE) ? 8 : 16);
	printf ("Erase is %s\n", (erase == TRUE) ? "ON" : "OFF");
	printf ("Verify is %s\n", (verify == TRUE) ? "ON" : "OFF");
#endif

	/* Setup location of flash control registers */
	if (Flash_Width == BDM_SIZE_WORD) {
		Flash_Reg1 = Flash_Base_Offset + Flash_Write_Offset + (0x05555<<1);
		Flash_Reg2 = Flash_Base_Offset + Flash_Write_Offset + (0x02AAA<<1);
	} else {
		Flash_Reg1 = Flash_Base_Offset + Flash_Write_Offset + 0x05555;
		Flash_Reg2 = Flash_Base_Offset + Flash_Write_Offset + 0x02AAA;
	}

	bdm_open();
	bdm_setdelay(10);
	hashmark = 1;
	if (*fname) bdm_do_load_macro(fname,BEGIN_MACRO);
	bdm_setdelay(1);
	done = cmd = 0;

	while (!done) {
		if (erase == TRUE) {
			runerase();
			erase = FALSE;
		}

	        if (load == TRUE) {
		       bdm_load(fname);
		       load = FALSE;
		}
		else {
		       bdm_reset();
                       if (!(bdm_get_status() & BDM_TARGETSTOPPED))
                               bdm_stop_chip();

                       bdm_do_load_macro(fname,BEGIN_MACRO);
		}

		printf("bdm-load %d> ",++cmd);
                if (fgets(buf,sizeof(buf),stdin)) {
char *s=strchr(buf,'\n');
			if (s) *s='\0';
			}
		else {
			*buf='\0';
			done=1;
			}

		if (!strcmp(buf,"run")) {
			bdm_set_sys_reg(BDM_REG_RPC,entry_pt);
			bdm_go();
		}

		if (!strcmp(buf,"reset")) {
			bdm_init();
			bdm_reset();
		}

		if (!strncmp(buf,"erase", (strlen (buf) < 5) ? strlen (buf) : 5))
			erase = TRUE;

		if (!strncmp(buf,"load",4)) {
			load = TRUE;
			if ((strlen(buf) > 5)) {
				strcpy(fname,&buf[5]);
			}
		}

		if (!strcmp(buf,"exit") || !strcmp(buf,"quit")) done = 1;

		if (!strncmp(buf,"make",4)) {
			system (buf);
			load = TRUE;
		}
	}
	exit (0);
}

